#include <SoftwareSerial.h>
#include <SolarCalculator.h>

#define GPS
//#define LCD
#define OLED_WIRE
//#define OLED_AVRI2c

#ifdef GPS
#include <TinyGPSPlus.h>
static const int RXPin = 4, TXPin = 3;
static const uint32_t GPSBaud = 9600;
TinyGPSPlus gps;
SoftwareSerial gpsSerial(RXPin, TXPin);
#endif

#ifdef LCD
#include "LiquidCrystal_I2C.h"
LiquidCrystal_I2C lcd(0x27, 20, 4);
#endif

#ifdef OLED_AVRI2C
#include "SSD1306Ascii.h"
#include "SSD1306AsciiAvrI2c.h"
#define I2C_ADDRESS 0x3C
#define RST_PIN -1
SSD1306AsciiAvrI2c oled;
#define OLED
#endif

#ifdef OLED_WIRE
#include "SSD1306Ascii.h"
#include "SSD1306AsciiWire.h"
#define I2C_ADDRESS 0x3C
#define RST_PIN -1
SSD1306AsciiWire oled;
#define OLED
#endif

static const int Relay_out = 6;
static const int Led_GPS = 7;
int nb_blink = 0;

double offset_lever = -10;
double offset_coucher = 30;  // offset en minutes
char str[6];

// Rounded HH:mm format
char *hoursToString(double h, char *str) {
  int m = int(round(h * 60));
  int hr = (m / 60) % 24;
  int mn = m % 60;

  str[0] = (hr / 10) % 10 + '0';
  str[1] = (hr % 10) + '0';
  str[2] = ':';
  str[3] = (mn / 10) % 10 + '0';
  str[4] = (mn % 10) + '0';
  str[5] = '\0';
  return str;
}

int year = 2022;
int month = 01;
int day = 01;
int hour = 12;
int minute = 0;
int second = 0;

byte last_second;
bool bool_jour = 0;
int nb_sat = 0;
//TinyGPSCustom sats(gps, "GPGSV", 3);

// Coordonnées par défaut
double latitude = 46.723946; // Centre France
double longitude = 2.499648;

double transit = 0, sunrise = 0, sunset = 0;

#ifdef LCD
void LCDPrintDigits(int digits) {
  if (digits < 10)
    lcd.print('0');
  lcd.print(digits);
}

void affichage_LCD() {
  lcd.setCursor(0, 0);
  LCDPrintDigits(hour);
  lcd.print(":");
  LCDPrintDigits(minute);
  lcd.print(":");
  LCDPrintDigits(second);
  lcd.setCursor(10, 0);
  LCDPrintDigits(day);
  lcd.print("/");
  LCDPrintDigits(month);
  lcd.print("/");
  lcd.print(year);
  lcd.setCursor(0, 1);
  lcd.print(latitude, 4);
  lcd.setCursor(14, 1);
  lcd.print(longitude, 4);
  lcd.setCursor(0, 2);
  lcd.print("Lever: ");
  lcd.setCursor(9, 2);
  lcd.print(hoursToString((sunrise + (offset_lever / 60)), str));
  lcd.setCursor(0, 3);
  lcd.print("Coucher: ");
  lcd.setCursor(9, 3);
  lcd.print(hoursToString((sunset + (offset_coucher / 60)), str));
  if (bool_jour == 1) {
    lcd.setCursor(15, 3);
    lcd.print(F("Jour"));
  } else {
    lcd.setCursor(15, 3);
    lcd.print(F("Nuit"));
  }
}
#endif

#ifdef OLED
void oledPrintDigits(int digits) {
  if (digits < 10)
    oled.print('0');
  oled.print(digits);
}

void affichage_oled() {
  oled.setCursor(5, 1);
  oledPrintDigits(hour);
  oled.print(":");
  oledPrintDigits(minute);
  oled.print(":");
  oledPrintDigits(second);
  oled.setCursor(65, 1);
  oledPrintDigits(day);
  oled.print("-");
  oledPrintDigits(month);
  oled.print("-");
  oled.print(year);
  oled.setCursor(5, 3);
  oled.print(latitude, 4);
  oled.setCursor(70, 3);
  oled.print(longitude, 4);
  oled.setCursor(5, 5);
  oled.println("Ouverture");
  oled.setCursor(65, 5);
  oled.println(hoursToString((sunrise + (offset_lever / 60)), str));
  oled.setCursor(5, 7);
  oled.println("Fermeture");
  oled.setCursor(65, 7);
  oled.println(hoursToString((sunset + (offset_coucher / 60)), str));
  if (bool_jour == 1) {
    oled.setCursor(100, 5);
    oled.print(F("Jour"));
  } else {
    oled.setCursor(100, 5);
    oled.print(F("Nuit"));
  }
  oled.setCursor(100, 7);
  oled.print(F("St "));
  oled.setCursor(116, 7);  
  oledPrintDigits(nb_sat);
}
#endif

void calcul_jour_nuit() {
  double maintenant = (hour + (minute / 60.0));
  if (((sunrise + (offset_lever / 60)) <= maintenant) && (maintenant < (sunset + (offset_coucher / 60))) == 1) {
    bool_jour = 1;
    digitalWrite(Relay_out, HIGH);
  } else {
    bool_jour = 0;
    digitalWrite(Relay_out, LOW);
  }
}

void setup() {
  Serial.begin(9600);

#ifdef OLED_WIRE
  Wire.begin();
  Wire.setClock(400000L);
#endif

  pinMode(Relay_out, OUTPUT);
  digitalWrite(Relay_out, LOW);
  pinMode(Led_GPS, OUTPUT);
  digitalWrite(Led_GPS, LOW);

#ifdef GPS
  gpsSerial.begin(GPSBaud);
  SoftwareSerial gpsSerial(RXPin, TXPin);
#endif

#ifdef LCD
  lcd.init();
  lcd.backlight();
  lcd.setCursor(2, 0);
  lcd.print("Cyber Poulailler");
  lcd.setCursor(2, 2);
  lcd.print("F1TE-F5DJL  2022");
  delay(4000);
  lcd.clear();
#endif

#ifdef OLED
  //oled.begin(&Adafruit128x64, I2C_ADDRESS);
  oled.begin(&SH1106_128x64, I2C_ADDRESS);
  delay(1000);
  oled.setFont(Adafruit5x7);
  oled.clear();
  oled.setCursor(15, 2);
  oled.print("Cyber Poulailler");
  oled.setCursor(20, 5);
  oled.print("F1TE-F5DJL 2022");
  delay(4000);
  oled.clear();
#endif
}

#ifdef GPS

void getgps(TinyGPSPlus &gps) {

  nb_blink = 0;

  if (gps.date.isValid()) {
    year = gps.date.year();
    month = gps.date.month();
    day = gps.date.day();
  }
  if (gps.time.isValid()) {
    hour = gps.time.hour();
    minute = gps.time.minute();
    second = gps.time.second();
    nb_blink = 1;
  }
  if (gps.location.isValid()) {
    latitude = gps.location.lat();
    longitude = gps.location.lng();
    nb_blink = nb_blink + 1;
  }
  if (gps.satellites.isValid()) {
    nb_sat = gps.satellites.value();
  }
}
#endif

static void smartDelay(unsigned long ms) {
  unsigned long start = millis();
  do {
#ifdef GPS
    while (gpsSerial.available())
      gps.encode(gpsSerial.read());
#endif
  } while (millis() - start < ms);
}

void blink_led(int nb_blink) {
  if (nb_blink >= 0) {
    digitalWrite(Led_GPS, 1);
    delay(75);
    digitalWrite(Led_GPS, 0);
    delay(75);
  }
  if (nb_blink == 2) {
    digitalWrite(Led_GPS, 1);
    delay(75);
    digitalWrite(Led_GPS, 0);
  }
}

void loop() {
  smartDelay(700);

#ifdef GPS
  getgps(gps);
#endif
  blink_led(nb_blink);
  calcSunriseSunset(year, month, day, latitude, longitude, transit, sunrise, sunset);
  calcul_jour_nuit();
#ifdef OLED
  affichage_oled();
#endif
#ifdef LCD
  affichage_LCD();
#endif
}
